/*
   mergelogs.c

   Copyright (C) 2001  Ulric Eriksson <ulric@siag.nu>
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>

#define KEEP_MAX 100	/* how much to keep from the URI */

typedef struct {
	char *a;	/* server name */
	char *fn;	/* log file name */
	FILE *fp;
	char **p;	/* buffered lines from log file */
	int np;		/* number of buffered lines */
	int hw;		/* high water mark */
} server;

static server *servers;
static int nservers;
static int debuglevel = 0;

static char *pfile;
static int jitter = 2;
static int lag = 0;

static char *months[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

static void debug(char *fmt, ...)
{
        va_list ap;
        va_start(ap, fmt);
        vfprintf(stderr, fmt, ap);
        va_end(ap);
        fprintf(stderr, "\n");
}

static void error(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
	exit(1);
}

static char *copy_string(char *p)
{
	char *q = malloc(strlen(p)+1);
	if (q == NULL) error("Can't allocate memory for string");
	strcpy(q, p);
	return q;
}

static void usage(void)
{
	printf("Usage: mergelogs -p penlog [-j jitter] [-l lag] \\\n"
	       "        server1:logfile1 [server2:logfile2 ...]\n\n"
	       "  -d              Debugging (repeat for more)\n"
	       "  -p penlog       Log file from pen\n"
	       "  -j jitter       Jitter in seconds [2]\n"
	       "  -l lag          Lag in seconds [0]\n"
	       "  server:logfile  Web server address and name of logfile\n");
	exit(0);
}

static int options(int argc, char **argv)
{
	int c;

	while ((c = getopt(argc, argv, "p:a:j:l:d")) != EOF) {
		switch (c) {
		case 'p':
			pfile = optarg;
			break;
		case 'j':
			jitter = atoi(optarg);
			break;
		case 'l':
			lag = atoi(optarg);
			break;
		case 'd':
			debuglevel++;
			break;
		default:
			usage();
		}
	}
	return optind;
}

static int mon2num(char *p)
{
	int i;

	for (i = 0; i < 12; i++)
		if (!strcmp(p, months[i])) return i;
	return -1;
}

static time_t whenp(char *p, struct tm *tms)
{
	char dd[100], mm[100], yy[100];
	char hh[100], mi[100], ss[100], tz[100];
	time_t t;

	sscanf(p, "%[^/]/%[^/]/%[^:]:%[^:]:%[^:]:%s %s",
		dd, mm, yy, hh, mi, ss, tz);
	tms->tm_sec = atoi(ss);
	tms->tm_min = atoi(mi);
	tms->tm_hour = atoi(hh);
	tms->tm_mday = atoi(dd);
	tms->tm_mon = mon2num(mm);
	tms->tm_year = atoi(yy)-1900;
	t = mktime(tms);
	return t;
}

static time_t logtime(char *p)
{
	char b[1024], d[1024];
	struct tm tms;

	if (sscanf(p, "%s %s %s [%[^]]]", d, d, d, b) != 4) return 0;
	else return whenp(b, &tms);
}

static char *loguri(char *p)
{
	static char b[1024];
	char d[1024];
	if (sscanf(p, "%s %s %s [%[^]]] \"%[^\"]", d, d, d, d, b) != 5)
		return NULL;
	return b;
}

/* Return the index of the server with a given IP address, or -1 for failure */
static int server_a2i(char *a)
{
	int i;

	for (i = 0; i < nservers; i++) {
		if (!strcmp(servers[i].a, a)) return i;
	}
	return -1;
}

static int too_old(long t1, long t2)
{
	return t1 < t2+lag-jitter;
}

static int too_new(long t1, long t2)
{
	return t1 > t2+lag+jitter;
}

/* We want to cache lines from the web server log that are no older
   than (when+lag-jitter) but no newer than (when+lag+jitter)
*/
static void update_server(int i, long when)
{
	int j, k;
	char b[1024];
	long then;

	/* Remove lines that are too old */
	for (j = 0; j < servers[i].np; j++) {
		then = logtime(servers[i].p[j]);
		if (too_old(then, when)) {
			if (debuglevel) debug("%ld too old, removing %d.",
					      then, j);
			free(servers[i].p[j]);
			for (k = j+1; k < servers[i].np; k++) {
				servers[i].p[k-1] = servers[i].p[k];
			}
			servers[i].np--;
		}
	}
	/* Add lines until EOF or we get one that is too new */
	for (;;) {
		if (feof(servers[i].fp)) break;
		if (fgets(b, sizeof b, servers[i].fp) == NULL) break;
		then = logtime(b);
		if (debuglevel > 1) debug("then = %ld, when = %ld",
					  then, when);
		if (too_old(then, when)) continue;
		/* We want this one */
		if (servers[i].np >= servers[i].hw) {
			servers[i].hw = 2*servers[i].np+1;
			servers[i].p = realloc(servers[i].p, servers[i].hw);
		}
		servers[i].p[servers[i].np++] = copy_string(b);
		if (too_new(then, when)) break;
	}
	if (debuglevel) debug("%d lines", servers[i].np);
}

static int best_match(int i, long when, char *uri)
{
	int j, best = -1;
	long then;

	for (j = 0; j < servers[i].np; j++) {
		then = logtime(servers[i].p[j]);
		if (debuglevel > 1) debug("when = %ld, then = %ld",
					  when, then);
		if (too_old(then, when)) continue;
		if (too_new(then, when)) continue;
		if (!strncmp(uri, loguri(servers[i].p[j]), KEEP_MAX)) best = j;
	}
	return best;
}

static void rewrite(char *from, char *line)
{
	char b[1024], d[1024];
	if (sscanf(line, "%s %[^\n]", d, b) != 2) {
		if (debuglevel) debug("Bogus logline '%s'", line);
		return;
	}
	printf("%s %s\n", from, b);
}

int main(int argc, char **argv)
{
	char b[1024], from[1024], to[1024], uri[1024];
	int i, n, s;
	long when;
	FILE *pfp = 0;

	n = options(argc, argv);
	argc -= n;
	argv += n;
	if (argc < 1) {
		usage();
	}
	if (pfile == NULL) error("pfile null");
	s = 0;
	servers = calloc(argc, sizeof *servers);
	for (i = 0; i < argc; i++) {
		servers[s].a = copy_string(argv[i]);
		servers[s].fn = strchr(servers[s].a, ':');
		if (servers[s].fn == NULL)
			error("Bogus server '%s'\n", argv[i]);
		*servers[s].fn++ = '\0';
		servers[s].fp = fopen(servers[s].fn, "r");
		if (servers[s].fp == NULL)
			error("Can't open logfile '%s'\n", servers[s].fn);
		servers[s].p = NULL;
		servers[s].np = 0;
		servers[s].hw = 0;
		nservers++;
		s++;
	}
	if (pfile) pfp = fopen(pfile, "r");
	else error("No pen logfile");
	if (pfp == NULL) error("pfp null");

	if (debuglevel) debug("Pen logfile records:");
	while (fgets(b, sizeof b, pfp)) {
		if (sscanf(b, "%s %ld %s %[^\n]", from, &when, to, uri) == 4) {
			if (debuglevel) {
				debug("From: %s", from);
				debug("To: %s", to);
				debug("When: %ld", when);
				debug("What: %s", uri);
			}
			i = server_a2i(to);
			if (debuglevel) debug("Server: %d", i);
			if (i == -1) continue;
			update_server(i, when);
			s = best_match(i, when, uri);
			if (s == -1) {
				if (debuglevel)
					debug("No match, discarding");
			} else {
				if (debuglevel) {
					debug("Best match: %s",
					      servers[i].p[s]);
				}
				rewrite(from, servers[i].p[s]);
			}
		} else {
			if (debuglevel) debug("Bogus log line '%s'", b);
		}
	}
	for (i = 0; i < nservers; i++) {
		fclose(servers[i].fp);
	}
	fclose(pfp);
	return 0;
}
